@@ -86,6 +86,9 @@ public class JLHAFrontEnd {
 	String[] filesFilename=null;
 	int i;
 
+	/* set file encoding to ISO8859-1 as default */
+	// java.util.Locale.setDefault(java.util.Locale.US); /* FIXME: ugly, but work */
+
 	/* check args */
 	if (args.length < 2 && args.length != 1) {
 	    usage(args);
@@ -188,76 +191,6 @@ public class JLHAFrontEnd {
 	return i;
     }
 
-    /**
-     * Transfer an InputStream to an OutputStream
-     *
-     * @param in the input stream
-     * @param out the output stream
-     * @param ent the distribution array
-     * @return number of blocks (8k bytes) copied
-     *
-     */
-    private int transferTo(InputStream in, OutputStream out, long[] ent) {
-	int i,j,bufRead,k;
-	byte[] buf = new byte[1024*8];
-	for (j=0 ; j<256 ; j++) {
-	    ent[j]=0;
-        }
-	for (i=0; true; i++) {
-	    try {
-		bufRead=in.read(buf);
-	    } catch (Exception e) {
-		mylogger.severe("read from in "+e.toString());
-		break;
-	    }
-	    if (bufRead<=0) break;
-	    for (j=0 ; j<bufRead ; j++) {
-	        k=buf[j];
-	        while (k<0) {
-	            k+=256;
-                }
-                k = k%256;
-                ent[k]++;
-            }
-	    try {
-		out.write(buf,0,bufRead);
-	    } catch (Exception e) {
-		mylogger.severe("write to out "+e.toString());
-		break;
-	    }
-	}
-	return i;
-    }
-
-    /**
-     * calculate entropy length
-     *
-     * @param ent distribution table
-     * @return estimate compreesed length
-     */
-    private long calEntropy(long[] ent) {
-        long N=0,ret=0;
-        double Hs=0.0,dN,dA,dRet;
-        int i;
-        for (i=0 ; i<ent.length ; i++) {
-            N += ent[i];
-        }
-        if (N<=0) {
-            return 0;
-        }
-        dN = (double)N;
-        for (i=0 ; i<ent.length ; i++) {
-          if (ent[i]==0) {
-              continue;
-          }
-          dA = ((double)(ent[i]));
-          Hs += (dA/dN)*(Math.log(dN)-Math.log(dA))/(Math.log(256));
-        }
-        dRet = N*(Hs);
-        ret = (long)dRet;
-        return ret;
-    }
-
     /**
      * transfer files in backup file to new archive file
      *
@@ -350,89 +283,6 @@ public class JLHAFrontEnd {
 	return bakfilename;
     }
 
-    /**
-     * add a file separator after a string if it is not 
-     * ended with the file separator 
-     *
-     * @param a original string
-     * @return new string with separator in the end
-     */
-    String addFileSeparator(String a) {
-	String ret;
-	if (a==null) {
-	    return a;
-	}
-	if (a.length()==0) {
-	    return File.separator;
-	}
-	if (a.charAt(a.length()-1)!=File.separatorChar) {
-	    ret = new String(a + File.separator);
-	} else {
-	    ret = new String(a);
-        }
-	return ret;
-    }
-
-    /**
-     * generate a list of files by the given files/directories
-     * 
-     * @param filename the given files/directories
-     * @return the list of the files
-     */
-    String[] genListArray(String[] filename) {
-	Vector data = new Vector();
-	String[] ret=null;
-	int i,n;
-	for (i=0 ; i<filename.length ; i++) {
-	    genListArray(filename[i],data);
-	}
-	n = data.size();
-	if (n>0) {
-	    ret = new String[n];
-	    for (i=0 ; i<n ; i++) {
-		ret[i] = (String)data.elementAt(i);
-	    }
-	}
-	return ret;
-    }
-
-    /**
-     * generate a list of files by the given file/directory
-     * 
-     * @param filename the given file/directory
-     * @param data return value, the list of the files
-     */
-    void genListArray(String filename,Vector data) {
-	File fl=null;
-	File[] childs=null;
-	int i;
-	String dirname=null;
-
-        try {
-	    fl = new File(filename);
-	} catch (Exception e) {
-	    mylogger.info("open file error");
-	    return;
-	}
-
-	try {
-	    if (fl.isFile()) {
-		data.add(filename);
-	    } else if (fl.isDirectory()) {
-		dirname = addFileSeparator(filename);
-		data.add(dirname);
-		childs = fl.listFiles();
-		for (i=0 ; i<childs.length ; i++) {
-		    genListArray(dirname+childs[i].getName(),data);
-		}
-	    } else {
-		mylogger.severe("fl is unknown");
-	    }
-	} catch (SecurityException e) {
-	    mylogger.severe("operation error "+e.toString());
-	}
-    }
-
     /**
      * execute add command
      *
@@ -446,7 +296,8 @@ public class JLHAFrontEnd {
 	File fl=null;
 	String backupFilename=null;
 	LhaFile backupFile=null;
-	LhaOutputStream lio=null;
+	LhaWithEntropyOutputStream lio=null;
+        FileList fileList=null;
 
 	/* open archive file */
 	try {
@@ -473,14 +324,15 @@ public class JLHAFrontEnd {
 
 	/* open new archive file */
 	try {
-	    lio = new LhaOutputStream(new FileOutputStream(archiveFilename));
+	    lio = new LhaWithEntropyOutputStream(new FileOutputStream(archiveFilename));
 	} catch (Exception e) {
 	    mylogger.severe("new LhaImeediateOutputStream with "+archiveFilename+" error: "+e.toString());
 	    return;
 	}
 
 	/* generate list */
-	list = genListArray(filesFilename);
+	fileList = new FileList(filesFilename);
+	list = fileList.getFiles();
 
 	/* open backup file */
 	try {
@@ -528,8 +380,8 @@ public class JLHAFrontEnd {
 	String backupFilename=null;
 	LhaFile backupFile=null;
 	LhaHeader[] backupFileEntries=null;
-	LhaOutputStream lio=null;
-	Vector newlist;
+	LhaWithEntropyOutputStream lio=null;
+	FileList fileList=null;
 
 	/* open archive file */
 	try {
@@ -556,14 +408,15 @@ public class JLHAFrontEnd {
 
 	/* open new archive file */
 	try {
-	    lio = new LhaOutputStream(new FileOutputStream(archiveFilename));
+	    lio = new LhaWithEntropyOutputStream(new FileOutputStream(archiveFilename));
 	} catch (Exception e) {
 	    mylogger.severe("new LhaImeediateOutputStream with "+archiveFilename+" error: "+e.toString());
 	    return;
 	}
 
 	/* generate list */
-	list = genListArray(filesFilename);
+	fileList = new FileList(filesFilename);
+	list = fileList.getFiles();
 
 	/* open backup file */
 	try {
@@ -579,8 +432,7 @@ public class JLHAFrontEnd {
 
 	/* compare createtime */
 	backupFileEntries = backupFile.getEntries();
-	newlist = new Vector();
-	for (i=0 ; i<list.length ; i++) {
+	for (i=0 ; list != null && i<list.length ; i++) {
 	    File f1;
 	    long modtime,oldmodtime;
 
@@ -600,7 +452,7 @@ public class JLHAFrontEnd {
 	    }
 	    
 	    if (modtime==0) {
-		newlist.add(list[i]);
+	        fileList.setFileMark(list[i]);
 		continue;
 	    }
 
@@ -612,22 +464,21 @@ public class JLHAFrontEnd {
 	    }
 
 	    if (oldmodtime == 0 || modtime > oldmodtime) {
-		newlist.add(list[i]);
-	    }
+	        fileList.setFileMark(list[i]);
+            }
 	    
 	}
 
-	list = new String[newlist.size()];
-	for (i=0 ; i<newlist.size() ; i++) {
-	    list[i] = (String)newlist.elementAt(i);
-	}
+	list = fileList.getMarkedFiles();
 
 	transferBackupToArchive(backupFile,lio,list);
 
 	/* compress new entries */
-	for (i=0 ; i<list.length ; i++) {
-	    createSwitcher(useropts,lio,list[i]);
-	}
+	if (list != null) {
+	  for (i=0 ; i<list.length ; i++) {
+	      createSwitcher(useropts,lio,list[i]);
+	  }
+        }
 	try {
 	    lio.close();
 	} catch (IOException e) {
@@ -845,7 +696,7 @@ public class JLHAFrontEnd {
 	    /* add working directory */
 	    if (workingDir != null) {
 	        File tmpfile=null;
-		filename = addFileSeparator(workingDir)+filename;
+		filename = FileList.addFileSeparator(workingDir)+filename;
 		tmpfile = new File(workingDir);
 		if (tmpfile != null) {
 		  tmpfile.mkdirs();
@@ -1359,7 +1210,7 @@ public class JLHAFrontEnd {
      * @param lio the LhaOutputStream for an archive file
      * @param filename the file need to be compressed
      */
-    public void createFile(Opts useropts, LhaOutputStream lio, String filename) {
+    public void createFile(Opts useropts, LhaWithEntropyOutputStream lio, String filename) {
 	LhaHeader header = new LhaHeader( filename );
 	FileInputStream raf = null;
 	File fl;
@@ -1368,7 +1219,6 @@ public class JLHAFrontEnd {
 	ListFormatter lf = new ListFormatter();
 	int compressMethod=0;
 	long compressedSize=0;
-	long[] distributiontable=new long[256];
 	OutputStream liodata;
 
 	compressMethod = useropts.getCompressionMethod();
@@ -1388,7 +1238,9 @@ public class JLHAFrontEnd {
 	    mylogger.severe("new FileInputStream: "+e.toString());
 	    return;
 	}
-
+        if (useropts.getHeaderLevel() != -1) {
+	  header.setHeaderLevel(useropts.getHeaderLevel());
+        }
 	header.setLastModified( new Date( fl.lastModified() ) );
 	header.setOriginalSize( fl.length());
 	header.setOSID((byte)0);
@@ -1419,7 +1271,7 @@ public class JLHAFrontEnd {
 	if (useropts.getEUCMode()) {
 	    liodata = new EUCJP2SJISOutputStream(liodata);
 	}
-	j=transferTo(raf,liodata,distributiontable);
+	j=transferTo(raf,liodata);
 	try {
 	    lio.closeEntry();
 	} catch (IOException e) {
@@ -1427,7 +1279,7 @@ public class JLHAFrontEnd {
 	}
 	if (quiet==0) {
 	    //compressedSize = header.getCompressedSize();
-	    compressedSize = calEntropy(distributiontable); /* FIXME: display correct compression rate */
+	    compressedSize = lio.getEntropyLength(); /* FIXME: display correct compression rate */
 	    if (header.getOriginalSize() > 0 && header.getOriginalSize() > compressedSize) {
         	System.out.print(lf.toField("("+Long.toString(compressedSize*100/header.getOriginalSize())+"%)",6,true));
             } else {
@@ -1458,7 +1310,7 @@ public class JLHAFrontEnd {
      * @param lio the LhaOutputStream for an archive file
      * @param filename the directory need to be compressed
      */
-    public void createDirectory(Opts useropts,LhaOutputStream lio, String dirname) {
+    public void createDirectory(Opts useropts,LhaWithEntropyOutputStream lio, String dirname) {
 	File fl=null;
 	File[] listfiles=null;
 	String prefix=null;
@@ -1468,7 +1320,7 @@ public class JLHAFrontEnd {
 
 	deleteFiles=useropts.getDeleteFiles();
 
-	prefix = addFileSeparator(dirname);
+	prefix = FileList.addFileSeparator(dirname);
 
 	try {
 	    fl = new File(dirname);
@@ -1478,6 +1330,9 @@ public class JLHAFrontEnd {
 	}
 	/* add -lhd- record */
 	header = new LhaHeader(prefix);
+        if (useropts.getHeaderLevel() != -1) {
+	  header.setHeaderLevel(useropts.getHeaderLevel());
+        }
 	header.setCompressMethod(CompressMethod.LHD);
 	header.setLastModified( new Date( fl.lastModified() ) );
 	header.setOSID((byte)0);
@@ -1517,7 +1372,7 @@ public class JLHAFrontEnd {
      * @param lio the LhaOutputStream for an archive file
      * @param filename the filename
      */
-    public void createSwitcher(Opts useropts,LhaOutputStream lio, String filename) {
+    public void createSwitcher(Opts useropts,LhaWithEntropyOutputStream lio, String filename) {
 	File fl=null;
 	try {
 	    fl = new File(filename);
@@ -1557,14 +1412,14 @@ public class JLHAFrontEnd {
      * @param filesFilename the array of filenames to be processed
      */
     public void create(Opts useropts, String archiveFilename, String[] filesFilename) {
-	LhaOutputStream lio=null;
+	LhaWithEntropyOutputStream lio=null;
 	int i;
 	if (filesFilename == null || filesFilename.length==0) {
 	    System.err.println("LHa: Error: No files given in argument, do nothing.");
 	    return;
 	}
 	try {
-	    lio = new LhaOutputStream(new FileOutputStream(archiveFilename));
+	    lio = new LhaWithEntropyOutputStream(new FileOutputStream(archiveFilename));
 	} catch (Exception e) {
 	    mylogger.severe("new LhaImeediateOutputStream with "+archiveFilename+" error: "+e.toString());
 	    return;
@@ -1588,7 +1443,7 @@ public class JLHAFrontEnd {
 	int i,j,k;
 
 	String[] cmds = { "commands:", " a   Add to archive", " x,e Extract from archive", " l,v List / Verbose list", " u   Update newer files to archive", " d   Delete from archive", " m   Move to archive", " c   Create new archive", " p   Print to STDOUT from archive", " t   Test archive"  };
-	String[] opts = { "options:", " q{num} quiet mode", " v  verbose", " n  not execute" , " f  force", " t  FILES are TEXT files" , " o[567] compression method (a/u)", " w=<dir> specify working directory" ," d  delete files after (a/u/c)", " i  ignore directory path (x/e)", " z  files not compress (a/u)", " e  TEXT code convert from/to EUC-JP" , " y  filename multibyte convert"  };
+	String[] opts = { "options:", " q{num} quiet mode", " v  verbose", " n  not execute" , " f  force", " t  FILES are TEXT files" , " o[567] compression method (a/u)", " w=<dir> specify working directory" ," d  delete files after (a/u/c)", " i  ignore directory path (x/e)", " z  files not compress (a/u)", " 0/1/2  header level (a/u)", " e  TEXT code convert from/to EUC-JP" , " y  filename multibyte convert"  };
 	System.out.print(ListFormatter._toField("jLHA",27,true));
 	System.out.println("Copyright(C) 2002  Michel Ishizuka");
 	System.out.print(ListFormatter._toField("jlhafrontend",27,true));
